#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <wait.h>
#include <pthread.h>
#include <string.h>

#define bool int
#define true 1
#define false 0
#define print(...) do{printf(__VA_ARGS__); fflush(stdout); } while(0)

void reverse_string(char* str) {
    char * end = str + strlen(str) - 1;

    while (str < end) {
        *str ^= *end; *end ^= *str; *str ^= *end;
        str++;
        end--;
    }
}

void *reverser(void *param);

pthread_t* t_ids;
char **reverser_args;
pthread_mutex_t *starts;
pthread_mutex_t *ends;

void init(int nb) {
    //init starts
    starts = malloc(sizeof(pthread_mutex_t) * nb);
    for(int i=0; i<nb; i++) {
        pthread_mutex_init(&starts[i], NULL);
        pthread_mutex_lock(&starts[i]); //doivent commencer locked
    }

    //init args
    reverser_args = malloc(sizeof(char*)*nb);
    for(int i=0; i<nb; i++) reverser_args[i] = "";

    //init threads
    t_ids = malloc(sizeof(pthread_t)*nb);
    for(int i=0; i<nb; i++) {
        pthread_t tid;  //ID du thread
        pthread_attr_t attr;   // attributs du thread

        pthread_attr_init(&attr);   // initie les attributs
        print("PASSED ID:%d\n", i);

        long arg = i;    //important! LONG! pas INT! montrer int arg=i;
        pthread_create(&tid, &attr, reverser, (void *)arg); //une fois qu'on les a, creer les thread

        t_ids[i] = tid;
    }

    //init signals
    ends = malloc(sizeof(pthread_mutex_t) * nb);
    for(int i=0;i<nb;i++) pthread_mutex_init(&ends[i], NULL);
}

bool done;
void consumer(int nb_t, int nb_str, char **arr) {
    init(nb_t);
    done = false;

    srand(time(NULL));   // Initialization, should only be called once.

    int arr_size = 0;
    while (arr[arr_size] != NULL) arr_size++;

    char **consumer_1 = malloc(sizeof(char *) * nb_str);
    for (int i = 0; i < nb_str; i++) {
        char *str = arr[rand() % arr_size];
        consumer_1[i] = malloc(sizeof(char) * (1+strlen(str)));
        strcpy(consumer_1[i], str);

        int r_id = rand() % nb_t;
        pthread_mutex_lock(&(ends[r_id]));
        reverser_args[r_id] = consumer_1[i];
        pthread_mutex_unlock(&starts[r_id]);
    }

    print("DONE!\n");

    for(int i=0; i<nb_t; i++) pthread_mutex_lock(&ends[i]);

    done = true;    //equiv a pthread_kill
    for(int i=0; i<nb_t; i++) {
        pthread_mutex_unlock(&starts[i]);
        pthread_join(t_ids[i],NULL);
    }

    //for (int i = 0; i < nb_str; i++) print("%i: %s\n", i, consumer_1[i]);
}

void *reverser(void *param) {
    int my_id = (int) param;

    print("MY ID:%d\n", my_id);

    while (1) {
        pthread_mutex_lock(&starts[my_id]);
        if(done) break;

        reverse_string(reverser_args[my_id]);

        pthread_mutex_unlock(&ends[my_id]);
    }

    pthread_exit(0);
}

void single_threaded(int nb_str, char **arr) {
    int arr_size = 0;
    while (arr[arr_size] != NULL) arr_size++;

    char **consumer_1 = malloc(sizeof(char *) * nb_str);

    for(int i=0; i<nb_str; i++) {
        char *str = arr[rand() % arr_size];
        consumer_1[i] = malloc(sizeof(char) * (1+strlen(str)));
        strcpy(consumer_1[i], str);

        reverse_string(consumer_1[i]);
    }
    //for (int i = 0; i < nb_str; i++) print("%i: %s\n", i, consumer_1[i]);
}

int main() {
    char *arr[] = {"foo", "bar", "sup guys!", "vooodooodooodooo", NULL};

    int nb_str = 2000000;

    clock_t t0 = clock();   //c...
    consumer(1, nb_str, arr);
    double datetime_diff_ms = difftime(clock(), t0)/CLOCKS_PER_SEC;
    print("Threaded: %f\n", datetime_diff_ms);

    t0 = clock();
    single_threaded(nb_str, arr);
    datetime_diff_ms = difftime(clock(), t0)/CLOCKS_PER_SEC;
    print("Un threaded: %f\n", datetime_diff_ms);
    return 0;
}
